#ifndef __MUSIC_Hdef
#define __MUSIC_Hdef

#include "sizes.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef __GNUC__
 #define EMPTYARRAY 0
 #define pk __attribute__((packed))
#else
 #define EMPTYARRAY
 #define pk
#endif

typedef struct musicdriver
{
	const char *name;
	const char *longdescr;
	int active;
	int maxch;
	
	/* Key used for finding. Also useful for command line option char. */
	int optchar;
	
	/* Return 0 if ok, -1 if not */
	int (*detect)(void);
	void (*reset)(void);
	void (*close)(void);
	
	/* Volumes are in 0..63 range.      */
	/* Note is in midi format: no gaps. */
	void (*noteon)(int ch, int note, int volume);
	void (*noteoff)(int ch);
	void (*volumechange)(int ch, int volume);
	
	void (*patch)(int ch,
				  int noteadd, int midipatch,
				  int c4spd, BYTE *adlibpatch);
				  
	void (*setpan)(int ch, int value); /* value is -64..+64 (left..right) */
	
	void (*midictrl)(int ch, BYTE ctrl, BYTE value);
	void (*bend)(int ch, int value); /* value is -128..+127 (0=0, 64=+1 seminote) */
	
	void (*delay)(unsigned us);
	
	struct musicdriver *next;
} musicdriver;

extern void registerdrivers(musicdriver *driver, ...);
extern musicdriver *finddriver(int optchar);
extern void foreachdriver(void (*rut)(const musicdriver *, void *), void *a);

typedef struct InternalHdr
{
    WORD OrdNum pk;      /* Number of orders, is even                */
                         /* If bit 8 (&256) is set, there exist no   */
                         /* notes between 126..253 and the 7th bit   */
                         /* of note value can be used to tell if     */
                         /* there is instrument number.              */
    WORD InsNum pk;      /* Number of instruments                    */
    WORD PatNum pk;      /* Number of patterns                       */
                         /* If bit 9 set, all instruments have       */
                         /* the insname[] filled with AsciiZ name.   */
    WORD Ins2Num pk;     /* Number of instruments - for checking!    */
    BYTE InSpeed pk;     /* Initial speed in 7 lowermost bits        */
                         /* If highest bit (&128) set, no logaritmic */
                         /* adlib volume conversion needs to be done */
                         /* when playing with midi device.           */
    BYTE InTempo pk;     /* Initial tempo                            */
    BYTE ChanCount pk;   /* Reliable only if PatNum bit 9 set        */
} InternalHdr; /* Size: 2+2+2+2+1+1+1 = 8+3 = 11 */

typedef struct InternalSample
{
	SBYTE FT pk;		/*finetune value         1         */
						/*  signed seminotes.              */
	BYTE D[11] pk;		/*Identificators         11        */
						/* - 12nd byte not needed.         */
						/* six high bits of D[8]           */
						/* define the midi volume          */
						/* scaling level                   */
						/* - default is 63.                */
						/* D[9] bits 5..2 have             */
						/* the automatical SDx             */
						/* adjust value.                   */
						/* - default is 0.                 */
                        /* D[9] bits 6-7 are               */
                        /* currently unused.               */
	BYTE Volume pk;		/*0..63                  1         */
						/* If bit 6 set,                   */
						/*  the instrument will be         */
						/*  played simultaneously          */
						/*  on two channels when           */
						/*  playing with a midi device.    */
						/*  To gain more volume.           */
						/* If bit 7 set,                   */
						/*  the finetune value             */
						/*  affects also FM.               */
						/*  (SMP->AME conversion)          */
	WORD C4Spd pk;		/*8363 = normal..?       2         */
	BYTE GM pk;			/*GM program number      1         */
    BYTE Bank;          /*Bank number, 0=normal. 1         */
                        /*High 2 bits (6-7) are free.      */
	char insname[EMPTYARRAY];
	                    /*Only if PatNum&512 set           */
} InternalSample;       /*             total:16 = 0x10     */

typedef struct S3MHdr
{
    char Name[28] pk;
    char CtrlZ pk;
	char FileType pk;  /* 16 = S3M */
	char Filler1[2] pk;
	WORD OrdNum pk; /* Number of orders, is even                         */
	WORD InsNum pk; /* Number of instruments                             */
	WORD PatNum pk; /* Number of patterns                                */
	WORD Flags pk;  /* Don't care, nothing important.                    */
	WORD Version pk;/* ST3.20 = 0x1320, ST3.01 = 0x1301                  */
	WORD SmpForm pk;/* 2 = Unsigned samples, don't care, ei adlibin asia */
	char Ident[4] pk;/*"SCRM"                                            */
	BYTE GVol pk;   /* Global volume, don't care, ei vaikuta adlibbeihin */
	BYTE InSpeed pk;/* Initial speed                                     */
					/*   &128 = linearmidivol                            */
	BYTE InTempo pk;/* Initial tempo                                     */
	BYTE MVol pk;   /* Master volume, don't care, ei vaikuta adlibbeihin */
	BYTE UC pk;     /* Ultraclick removal, don't care, gus-homma         */
	BYTE DP pk;     /* Default channel pan, don't care, gussiin sekin    */
	char Filler2[10] pk;
	BYTE Channels[32] pk; /* Channel settings, 16..31 = adl channels */
} S3MHdr;

typedef struct ADLSample
{
	BYTE Typi pk;       /*2 = AMel           1           */
	char DosName[13] pk;/*AsciiZ file name   13          */
						/* Temporary used by intgen:     */
						/*  [0]=GM (unsigned)            */
						/*  [1]=ft (signed)              */
						/*  [2]=doublechannel flag       */
	SWORD ManualScale pk;/* This is actually part of the  */
						/* filename, but this is now for */
						/* intgen temporary.             */
	BYTE D[12] pk;      /*Identificators     12          */
	BYTE Volume pk;     /*                   1           */
	char Filler1[3] pk; /*                   3           */
	WORD C4Spd pk;      /*8363 = normal..?   2           */
	char Filler2[14] pk;/* May contain:      14          */
						/*  [2]=GM (unsigned)            */
						/*  [3]=ft (signed)              */
	char Name[28] pk;   /*AsciiZ sample name 28          */
	char Ident[4] pk;   /*"SCRI"             4           */
} ADLSample;             /*                   80 = 0x50   */

typedef struct Pattern
{
	WORD Len pk;
    BYTE *Ptr pk;
} Pattern;

#undef pk

#ifdef __cplusplus
}
#endif
#endif /* __MUSIC_Hdef */
